//
//  cbDecoder.h
//
//  Created by Ken Miyachi on 11/12/07.
//  Copyright 2011 B-Core.Inc . All rights reserved.
//

#ifndef cbDecoder_h
#define cbDecoder_h
#include <stdio.h>
#include "cbtypes.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef void*	(*mwcbMalloc)				(size_t size);
typedef void	(*mwcbFree)					(void* ptr);
    
///---------------mwcbDecoderGetWorkSize--------------///
///   Get work memory size
///   [Params]
///   int width  : input image size width 
///   int height : input image size height 
size_t mwcbDecoderGetWorkSize(int width, int height);    

///---------------mwcbDecoderInitialize--------------///
///   Decoder initialize, Set allocation func pointer
///   [Params]
///   mwcbMalloc malloc  : function pointer of malloc 
///   mwcbFree free      : function pointer of free  
void mwcbDecoderInitialize(mwcbMalloc malloc, mwcbFree free);

///---------------mwcbDecoderSetParams--------------///
///   Set decoder params
///   [Params]
///   int image_correction_param  : MW_CB_DECODER_USE_IMAGE_CORRECTION / MW_CB_DECODER_NONUSE_IMAGE_CORRECTION
///   int ec_level                : error correction param, MW_CB_CODING_CORRECTION_LEVEL_0 ~ 2
///   [Default]
///   int image_correction_param  = MW_CB_DECODER_NONUSE_IMAGE_CORRECTION
///   int ec_level                = MW_CB_CODING_CORRECTION_LEVEL_0
void mwcbDecoderSetParams(int image_correction_param, int ec_level);

    
///---------------mwcbDecoderSetColorLocation--------------///
///   Set the color pick up coordinate point 
///   [Params]
///   int location_table[15][2]   : absolute coordinate point for input image
///   [Default]
///   For 64x64 Image
///   int location_table[15][2] = {{  8, 22 },{ 20, 20 },{ 32, 22 },{ 42, 20 },{ 52, 28 },{ 60, 22 },{ 54, 42 },
///                                { 40, 36 },{ 32, 42 },{ 26, 32 },{ 18, 32 },{  4, 42 },{ 14, 56 },{ 26, 54 },{ 44, 56 }} ;    
void mwcbDecoderSetColorLocation(int location_table[15][2]);

///---------------mwcbDecoderQueryImage--------------///
///   Decode one image, Input Image must have RGB888 Pixel Format.
///   Return the error code. return value < 0 : error code, return value >=0 : actually used error correction level 
///   [Params]
///   void* src_imageData         : input image data
///   int width                   : input image size width
///   int height                  : input image size height
///   unsigned long long* id      : return code id
///   unsigned char* color_seq    : return color seqence : R = 0, G = 1, B = 2, Miss = 255
int  mwcbDecoderQueryImage(void* src_imageData,int width, int height, unsigned long long* id, unsigned char* color_seq);

///---------------mwcbDecoderQueryImageEx--------------///
///   Decode one image, Input Image must have RGB888 Pixel Format.
///   Return the error code. return value < 0 : error code, return value >=0 : actually used error correction level
///   *Note : this function for debugging, you can check the result of color quantization with color_dump
///   [Params]
///   void* src_imageData         : input image data
///   int width                   : input image size width
///   int height                  : input image size height
///   unsigned long long* id      : return code id
///   unsigned char* color_seq    : return color seqence : R = 0, G = 1, B = 2, Miss = 255
///   unsigned char* color_dump   : return the result of color quantization, 
///                                 The pixel format is 8bit grayscale image : R = 50, G = 100, B = 150, BackBround = 0
///   [USAGE]
///   unsigned char* color_dump
///   color_dump = (unsigned char*)(malloc(sizeof(unsigned char) * width * height);
///   mwcbDecoderQueryImageEx(src_imageData, width, height, &id, color_seq, color_dump);
///   ----- use color_dump data -----
///   free(color_dump)
int  mwcbDecoderQueryImageEx(void* src_imageData,int width, int height, unsigned long long* id, unsigned char* color_seq, unsigned char* color_dump);

///---------------mwcbDecoderFinalize--------------///
///   CB Decoder Finalizee
void mwcbDecoderFinalize();
    

    
#ifdef __cplusplus
}
#endif


#ifdef __cplusplus

typedef mwcbMalloc		Malloc;
typedef mwcbFree		Free;

namespace mw { 
namespace cb {
    class Decoder{
        
        public:
        size_t GetWorkSize(int width, int height){
            return mwcbDecoderGetWorkSize(width, height);
        }
        
        void Initialize(Malloc malloc,Free free){
            return mwcbDecoderInitialize(malloc, free);
        }
    	
        void SetParams(IMAGE_CORRECTION im, EC_LEVEL el){
            
            return mwcbDecoderSetParams((int)im, (int)el);
        }
        
        void SetColorLocation(int location_table[15][2]){
            return mwcbDecoderSetColorLocation(location_table);
        }

        ERROR_CODE QueryImage(void* src_imageData,int width, int height, unsigned long long* id, unsigned char* color_seq){
            return (ERROR_CODE)mwcbDecoderQueryImage(src_imageData, width, height, id, color_seq);
        }
        
        void Finalize(){
            return mwcbDecoderFinalize();
        }
        
        ERROR_CODE QueryImageEx(void* src_imageData,int width, int height, unsigned long long* id, unsigned char* color_seq, unsigned char* color_dump){
            return (ERROR_CODE)mwcbDecoderQueryImageEx(src_imageData, width, height, id, color_seq, color_dump);
        }
        
    };

}
}
#endif

#endif
