﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     applet.cpp

  Copyright (C)2009-2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 43587 $
 *---------------------------------------------------------------------------*/

#include "applet.h"

namespace
{
    void RestoreGraphicSetting()
    {
        // GPU レジスタ設定の復帰
        // ただし、下記は DMPGL API を使用して設定した状態の復帰のみ
        // レジスタ設定コマンドは全て再発行する必要があります。
        nngxUpdateState(NN_GX_STATE_ALL);
        nngxValidateState(NN_GX_STATE_ALL,GL_TRUE);
    }
}

bool TransitionHandler::s_IsExitRequired = false;
volatile bool TransitionHandler::s_IsAfterWakeUp = false;
nn::os::LightEvent TransitionHandler::s_AwakeEvent;
nn::os::CriticalSection TransitionHandler::s_CriticalSection;
nn::os::CriticalSection TransitionHandler::s_CriticalSectionForSleep;

void TransitionHandler::Initialize()
{
    // スリープ関連のコールバック設定
    nn::applet::SetSleepQueryCallback(SleepQueryCallback, 0);
    nn::applet::SetAwakeCallback(AwakeCallback, 0);
//     nn::applet::SetSleepCanceledCallback(NULL, 0); // SleepCanceledCallback は使わないことを推奨します

    s_AwakeEvent.Initialize(true);
    s_CriticalSection.Initialize();
    s_CriticalSectionForSleep.Initialize();

    // 起床中は常に Signal 状態にしておく
    s_AwakeEvent.Signal();

    // アプレットライブラリに関する機能を有効にします。
    // 引数を false にしておくと、nn::applet::EnableSleep() を呼ぶまでは
    // 自動的にスリープ拒否を行います。
    nn::applet::Enable(false);

    // gx, snd, dsp 各ライブラリの Initialize 関数は applet::Enable() の後に呼ばれなければならない
    // 特に、gx の初期化である nngxInitialize 関数は、Enable() 直後の終了判定の後に行われるべきです。

    // ロゴ表示時に電源ボタンが押されたなど、既に終了条件が立っていた場合、ここでアプリを終了させる
    if ( nn::applet::IsExpectedToCloseApplication() )
    {
        s_IsExitRequired = true;
    }
}

void TransitionHandler::Finalize()
{
    s_AwakeEvent.Finalize();
    s_CriticalSection.Finalize();
    s_CriticalSectionForSleep.Finalize();

    nn::applet::SetSleepQueryCallback(NULL, 0);
    nn::applet::SetAwakeCallback(NULL, 0);
}

void TransitionHandler::EnableSleep()
{
    // スリープ要求に対する返答を有効にします。
    // 蓋の状態チェックを行い蓋が閉じられているならスリープ要求が発生します。
    nn::applet::EnableSleep(true);
}

void TransitionHandler::Process()
{
    // スリープからの起床時は、1フレーム分の描画完了を待ってから LCD を有効にする
    if ( TransitionHandler::s_IsAfterWakeUp )
    {
        nn::gx::StartLcdDisplay();
        TransitionHandler::s_IsAfterWakeUp = false;
    }

    // GPU のレジスタ設定の整合性を考慮した場合、
    // 各種遷移、スリープ要求への返答は描画終了後に行うべき

    // 本体スリープ判定
    if ( nn::applet::IsExpectedToReplySleepQuery() )
    {
        // 何らかの理由で SleepQuery に対して REJECT を返す場合は、ここで判定して
        // REJECT を返し、この関数を抜ける

        // ファイルシステム処理中はスリープをしない
        if ( TryLockForSleep() )
        {
            // 必要に応じて、ここでスリープ前処理を行う

            // スリープ許可直前に AwakeEvent をクリアしておく
            s_AwakeEvent.ClearSignal();

            // スリープを許可し、速やかにメインスレッドを止める
            nn::applet::ReplySleepQuery(nn::applet::REPLY_ACCEPT);
            s_AwakeEvent.Wait();

            // 必要に応じて、ここでスリープ復帰時の処理を行う

            UnlockForSleep();
        }
    }

    // 終了要求をチェック
    if ( nn::applet::IsExpectedToCloseApplication() )
    {
        s_IsExitRequired = true;
        return;
    }

    // アプリケーションの遷移処理を行う

    // HOMEボタン処理（グラフィックスライブラリが初期化済みでなければならない）
    if ( nn::applet::IsExpectedToProcessHomeButton() )
    {
        // ここでフラグなどを判定してブロックを抜けるようにすれば HOME ボタン禁止区間を実現できる
        // nn::applet::ClearHomeButtonState() による HOME ボタンフラグ削除も必要

        if ( TryLock() )
        {
            nn::applet::ProcessHomeButtonAndWait();

            Unlock();

            // HOME メニューから終了された場合やバッテリー不足といった要因により
            // アプリ終了と判断された場合、アプリのメインループを抜けて終了処理へ
            if ( nn::applet::IsExpectedToCloseApplication() )
            {
                // 描画権は渡されていない
                s_IsExitRequired = true;
                return;
            }

            // グラフィックの設定を戻す
            RestoreGraphicSetting();
        }
    }

    // 電源ボタン処理（グラフィックスライブラリが初期化済みでなければならない）
    if ( nn::applet::IsExpectedToProcessPowerButton() )
    {
        // 電源ボタン処理はすみやかに行い、アプリ終了前に行うべき処理
        // （セーブデータ処理など）は nn::applet::ProcessPowerButtonAndWait() から
        // かえってきた後に行ってください
        nn::applet::ProcessPowerButtonAndWait();

        // アプリ終了と判断された場合、アプリのメインループを抜けて終了処理へ
        if ( nn::applet::IsExpectedToCloseApplication() )
        {
            // 描画権は渡されていない
            s_IsExitRequired = true;
            return;
        }

        // グラフィックの設定を戻す
        RestoreGraphicSetting();
    }
}

bool TransitionHandler::IsExitRequired()
{
    return s_IsExitRequired;
}

void TransitionHandler::Lock()
{
    s_CriticalSection.Enter();
}

bool TransitionHandler::TryLock()
{
    return s_CriticalSection.TryEnter();
}

void TransitionHandler::Unlock()
{
    s_CriticalSection.Leave();
}

void TransitionHandler::LockForSleep()
{
    s_CriticalSectionForSleep.Enter();
}

bool TransitionHandler::TryLockForSleep()
{
    return s_CriticalSectionForSleep.TryEnter();
}

void TransitionHandler::UnlockForSleep()
{
    s_CriticalSectionForSleep.Leave();
}

/*------------------------------------------------------------------------*
    スリープ問い合わせ時に呼ばれるコールバック
 *------------------------------------------------------------------------*/
AppletQueryReply TransitionHandler::SleepQueryCallback( uptr arg )
{
    NN_UNUSED_VAR(arg);

    if ( !nn::applet::IsActive() )
    {
        // Inactive なとき、メインスレッドは applet::WaitForStarting() で停止状態であり、
        // 他のスレッドも停止状態になっているはず（アプリケーション側の実装依存）
        return nn::applet::REPLY_ACCEPT;
    }
    else
    {
        // メインスレッドで、必要な処理を終えてからスリープするので、
        // ここでは保留しておく
        return nn::applet::REPLY_LATER;
    }
}

/*------------------------------------------------------------------------*
    スリープ復帰時に呼ばれるコールバック
 *------------------------------------------------------------------------*/
void TransitionHandler::AwakeCallback( uptr arg )
{
    NN_UNUSED_VAR(arg);
    s_AwakeEvent.Signal();

    s_IsAfterWakeUp = true;
}
