﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     demo_MemoryManager.cpp

  Copyright (C)2009-2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 38331 $
 *---------------------------------------------------------------------------*/

#include "demo/Utility/demo_MemoryManager.h"

#define DEBUG_PRINT 0

namespace demo
{
    namespace
    {
        const int ALIGNMENT_SYSTEM_BUFFER     = 4;     // 規定のうち、最大のものを指定
        const int ALIGNMENT_VERTEX            = 4;     // GLfloat型の 4バイトアライン規定を考慮
        const int ALIGNMENT_TEXTURE           = 128;
        const int ALIGNMENT_RENDER_BUFFER     = 64;    // 24ビットフォーマットのデプスバッファ(D24)は96Byteアラインが必要
        const int ALIGNMENT_DISPLAY_BUFFER    = 16;
        const int ALIGNMENT_3D_COMMAND_BUFFER = 16;
    
        inline uptr MathRoundup(uptr x, int base)
        {
            return ((x) + ((base)-1)) & ~((base)-1);
        }
    }

    namespace detail
    {
        MemoryManager::MemoryManager() :
        m_Initialized(false),
        m_pStartAddrFcram(NULL), 
        m_CurrentAddrVramA(NULL),
        m_CurrentAddrVramB(NULL), 
        m_HeapOnFcram(),
        m_AllocatedBlockSize(0)
        {
        }

        MemoryManager::~MemoryManager() 
        {
            Finalize(); 
        }

        void MemoryManager::Initialize(const uptr fcramAddress, const size_t memorySize)
        {
            if (m_Initialized)
            {
                return;
            }
        
            m_pStartAddrFcram = fcramAddress;
            NN_TASSERT_(m_pStartAddrFcram != NULL);
            
            m_AllocatedBlockSize = memorySize;
            
            m_HeapOnFcram.Initialize(m_pStartAddrFcram, memorySize, nn::os::ALLOCATE_OPTION_LINEAR);

            m_CurrentAddrVramA = nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMA);
            m_CurrentAddrVramB = nn::gx::GetVramStartAddr(nn::gx::MEM_VRAMB);
            
        #if DEBUG_PRINT
            NN_TLOG_("[demo::MemoryManager] Start address on Main Memory (FCRAM): 0x%08X\n", m_pStartAddrFcram);
            NN_TLOG_("[demo::MemoryManager] Start address on VRAM-A: 0x%08X\n", m_CurrentAddrVramA);
            NN_TLOG_("[demo::MemoryManager] Start address on VRAM-B: 0x%08X\n", m_CurrentAddrVramB);
        #endif
            
            m_Initialized = true;
        }

        void MemoryManager::Finalize(void)
        {
            m_HeapOnFcram.Finalize();

            m_Initialized = false;
        }

        void MemoryManager::PrintFreeMemorySize(void)
        {
            NN_TLOG_("[demo::MemoryManager] Free memory size on Main Memory (FCRAM) is 0x%06X(max 0x%06X)\n",
                m_HeapOnFcram.GetTotalFreeSize(), m_AllocatedBlockSize);
            NN_TLOG_("[demo::MemoryManager] Free memory size on VRAM-A is 0x%06X(max 0x%06X)\n",
                nn::gx::GetVramEndAddr(nn::gx::MEM_VRAMA) - m_CurrentAddrVramA, nn::gx::GetVramSize(nn::gx::MEM_VRAMA));
            NN_TLOG_("[demo::MemoryManager] Free memory size on VRAM-B is 0x%06X(max 0x%06X)\n",
                nn::gx::GetVramEndAddr(nn::gx::MEM_VRAMB) - m_CurrentAddrVramB, nn::gx::GetVramSize(nn::gx::MEM_VRAMB));
        }
        
        void* MemoryManager::Allocate(GLenum area, GLenum aim, GLuint id, GLsizei size)
        {
            if (!m_Initialized)
            {
                NN_TPANIC_("Not initialized.\n");
            }
            
            if (size == 0)
                return 0;
            
            // VRAM にアロケートする場合は、簡単のために VRAM をスタックのように扱う
            // さらに、断片化には対処しない
            // 断片化に対処するには、確保成功時に area, aim, id といった確保したメモリに対する情報を
            // のちのデアロケートのために保持しておく必要があるはず
            
            int addrAlign = 8;
            void* resultAddr = NULL;

            // 各データの配置に対するアラインメントの制約を考慮
            switch (aim)
            {
                case NN_GX_MEM_SYSTEM:
                    addrAlign = ALIGNMENT_SYSTEM_BUFFER;
                    break;
                case NN_GX_MEM_TEXTURE:
                    addrAlign = ALIGNMENT_TEXTURE;
                    break;
                case NN_GX_MEM_VERTEXBUFFER:
                    addrAlign = ALIGNMENT_VERTEX;
                    break;
                case NN_GX_MEM_RENDERBUFFER:
                    addrAlign = ALIGNMENT_RENDER_BUFFER;
                    break;
                case NN_GX_MEM_DISPLAYBUFFER:
                    addrAlign = ALIGNMENT_DISPLAY_BUFFER;
                    break;
                case NN_GX_MEM_COMMANDBUFFER:
                    addrAlign = ALIGNMENT_3D_COMMAND_BUFFER;
                    break;
                default:
                    NN_TPANIC_("Invalid parameter.(0x%X)\n", aim);
                    break;
            }
            
            switch (area)
            {
            case NN_GX_MEM_FCRAM:
                // FCRAM の場合は SDK サポートの拡張ヒープから割り当てるため管理しない
                if ( (resultAddr = m_HeapOnFcram.Allocate(size, addrAlign)) == NULL)
                {
                    NN_TPANIC_("Lack of resource on Main Memory (FCRAM).\n");
                }
                break;
            case NN_GX_MEM_VRAMA:
                {
                    if (MathRoundup(m_CurrentAddrVramA, addrAlign) + size > nn::gx::GetVramEndAddr(nn::gx::MEM_VRAMA))
                    {
                        NN_TPANIC_("Lack of resource on VRAM-A.\n");
                    }

                    m_CurrentAddrVramA = MathRoundup(m_CurrentAddrVramA, addrAlign);
                    resultAddr = reinterpret_cast<void*>(m_CurrentAddrVramA);
                    m_CurrentAddrVramA += size;
                }
                break;
            case NN_GX_MEM_VRAMB:
                {
                    if (MathRoundup(m_CurrentAddrVramB, addrAlign) + size > nn::gx::GetVramEndAddr(nn::gx::MEM_VRAMB))
                    {
                        NN_TPANIC_("Lack of resource on VRAM-B.\n");
                    }

                    m_CurrentAddrVramB = MathRoundup(m_CurrentAddrVramB, addrAlign);
                    resultAddr = reinterpret_cast<void*>(m_CurrentAddrVramB);
                    m_CurrentAddrVramB += size;
                }
                break;
            default:
                // 通常の使用ではまずありえないため PANIC
                NN_TPANIC_("Invalid parameter. (0x%X)\n", area);
                break;
            }
            
        #if DEBUG_PRINT
            if (aim == NN_GX_MEM_SYSTEM)
            {
                NN_TLOG_("[demo::MemoryManager] NN_GX_MEM_SYSTEM is allocated at 0x%08X - 0x%08X on %X\n",
                    resultAddr,
                    reinterpret_cast<int>(resultAddr) + size - 1, area);
            }
            else if (aim == NN_GX_MEM_TEXTURE)
            {
                NN_TLOG_("[demo::MemoryManager] NN_GX_MEM_TEXTURE is allocated at 0x%08X - 0x%08X on %X\n",
                    resultAddr,
                    reinterpret_cast<int>(resultAddr) + size - 1, area);
            }
            else if (aim == NN_GX_MEM_VERTEXBUFFER)
            {
                NN_TLOG_("[demo::MemoryManager] NN_GX_MEM_VERTEXBUFFER is allocated at 0x%08X - 0x%08X on %X\n",
                    resultAddr,
                    reinterpret_cast<int>(resultAddr) + size - 1, area);
            }
            else if (aim == NN_GX_MEM_DISPLAYBUFFER)
            {
                NN_TLOG_("[demo::MemoryManager] NN_GX_MEM_DISPLAYBUFFER is allocated at 0x%08X - 0x%08X on %X\n",
                    resultAddr,
                    reinterpret_cast<int>(resultAddr) + size - 1, area);
            }
            else if (aim == NN_GX_MEM_RENDERBUFFER)
            {
                NN_TLOG_("[demo::MemoryManager] NN_GX_MEM_RENDERBUFFER is allocated at 0x%08X - 0x%08X on %X\n",
                    resultAddr,
                    reinterpret_cast<int>(resultAddr) + size - 1, area);
            }
            else if (aim == NN_GX_MEM_COMMANDBUFFER)
            {
                NN_TLOG_("[demo::MemoryManager] NN_GX_MEM_COMMANDBUFFER is allocated at 0x%08X - 0x%08X on %X\n",
                    resultAddr,
                    reinterpret_cast<int>(resultAddr) + size - 1, area);
            }
        #else            
            NN_UNUSED_VAR(id);
        #endif   // #if DEBUG_PRINT

            return resultAddr;
        }

        /* DMPGL用のメモリデアロケータ */
        void MemoryManager::Deallocate(GLenum area, GLenum aim, GLuint id, void* addr)
        {
            NN_UNUSED_VAR(aim);
            NN_UNUSED_VAR(id);
                        
            if (! m_Initialized)
            {
                NN_TPANIC_("Not initialized.\n");
            }
            
        #if DEBUG_PRINT
            NN_TLOG_("[demo::MemoryManager] (id %d) was deallocated at 0x%08X on %X\n", id, addr, area);
        #endif
            
            switch (area)
            {
            case NN_GX_MEM_FCRAM:
                m_HeapOnFcram.Free(addr);
                break;

            case NN_GX_MEM_VRAMA:
            case NN_GX_MEM_VRAMB:
                // VRAM 上のバッファは簡単のため解放処理を行わない
                break;

            default:
                NN_TPANIC_("Invalid parameter.\n");
                break;
            }
        }
    }
                   
    namespace
    {
        demo::detail::MemoryManager s_MemoryManager;
    }

    NN_WEAK_SYMBOL void InitializeMemoryManager(const uptr fcramAddress, const size_t memorySize)
    {
        s_MemoryManager.Initialize(fcramAddress, memorySize);
    }

    NN_WEAK_SYMBOL void FinalizeMemoryManager(void)
    {
        s_MemoryManager.Finalize();
    }

    NN_WEAK_SYMBOL void PrintMemoryManagerInfo(void)
    {
         s_MemoryManager.PrintFreeMemorySize();
    }

    NN_WEAK_SYMBOL void* GetAllocator(GLenum area, GLenum aim, GLuint id, GLsizei size)
    {
         return s_MemoryManager.Allocate(area, aim, id, size);
    }

    NN_WEAK_SYMBOL void GetDeallocator(GLenum area, GLenum aim, GLuint id, void* addr)
    {
         s_MemoryManager.Deallocate(area, aim, id, addr);
    }

    NN_WEAK_SYMBOL void* Alloc(size_t size)
    {
         return s_MemoryManager.Allocate(NN_GX_MEM_FCRAM, NN_GX_MEM_SYSTEM, 0, size);
    }

    NN_WEAK_SYMBOL void Free(void* ptr)
    {
         s_MemoryManager.Deallocate(NN_GX_MEM_FCRAM, NN_GX_MEM_SYSTEM, 0, ptr);
    }

}
