﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     GuiControlManager.cpp

  Copyright (C)2009-2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 40635 $
 *---------------------------------------------------------------------------*/

#include "GuiControlManager.h"

namespace scene
{

// コンストラクタです。
ControlManager::ControlManager()
{
    m_touchStatusOld.x     = 0;
    m_touchStatusOld.y     = 0;
    m_touchStatusOld.touch = false;

    m_activeId  = CONTROL_ID_UNUSED;
    m_pCallback = NULL;
    m_pParam    = NULL;
}

// デストラクタです。
ControlManager::~ControlManager()
{
    Clear();
}

// コントロールを追加します。
void ControlManager::Add(ControlBase* pControl)
{
    if (pControl)
    {
        if (pControl->GetId() != CONTROL_ID_UNUSED)
        {
            for (ControlList::iterator it = m_controlList.begin(); it != m_controlList.end(); it++)
            {
                // IDの重複チェック
                NN_ASSERT((*it)->GetId() != pControl->GetId());
            }
        }

        // コントロールマネージャを設定する
        pControl->SetControlManager(this);

        m_controlList.push_front(pControl);
    }
}

// コントロールをすべて削除します。
void ControlManager::Clear()
{
    for (ControlList::iterator it = m_controlList.begin(); it != m_controlList.end(); it++)
    {
        delete *it;
    }

    m_controlList.clear();
}

// コントロールを検索します。
ControlBase* ControlManager::FindBy(u32 id)
{
    if (id == CONTROL_ID_UNUSED)
    {
        return NULL;
    }

    for (ControlList::iterator it = m_controlList.begin(); it != m_controlList.end(); it++)
    {
        if ((*it)->GetId() == id)
        {
            return (*it);
        }
    }

    return NULL;
}

// コントロールリストを取得します。
ControlList* ControlManager::GetList()
{
    return &m_controlList;
}

// 更新処理を行います。
void ControlManager::Update()
{
    // 入力を取得する
    m_touchReader.ReadLatest(&m_touchStatus);

    // 現在の入力
    u16  nowX     = m_touchStatus.x;
    u16  nowY     = m_touchStatus.y;
    bool nowTouch = m_touchStatus.touch;
    // 1フレーム前の入力
    u16  prevX     = m_touchStatusOld.x;
    u16  prevY     = m_touchStatusOld.y;
    bool prevTouch = m_touchStatusOld.touch;

    for (ControlList::iterator it = m_controlList.begin(); it != m_controlList.end(); it++)
    {
        if ((*it)->GetId() == CONTROL_ID_UNUSED)
        {
            continue;
        }

        s32 x = (*it)->GetX();
        s32 y = (*it)->GetY();

        u32 width  = (*it)->GetWidth();
        u32 height = (*it)->GetHeight();

        // 現在の状態
        ControlState nowState = (*it)->GetState();

        // 無効・非表示時は何もしない
        if (nowState == CONTROL_STATE_DISABLED || nowState == CONTROL_STATE_HIDE)
        {
            continue;
        }

        // コントロール内にいる
        if (nowX >= x && nowX <= x + width && nowY >= y && nowY <= y + height)
        {
            // 前回タッチされていない＆今回タッチされた→ペンダウン
            if (!prevTouch && nowTouch && m_activeId == CONTROL_ID_UNUSED)
            {
                (*it)->OnPenDown();
                m_activeId = (*it)->GetId();

                if (m_pCallback)
                {
                    m_pCallback(CONTROL_EVENT_PENDOWN, *it, nowX - x, nowY - y, m_pParam);
                }
            }
            // 前回タッチされた＆今回もタッチされた→ペンスライド
            if (prevTouch && nowTouch && m_activeId == (*it)->GetId())
            {
                (*it)->OnPenSlide(true);

                if (m_pCallback)
                {
                    m_pCallback(CONTROL_EVENT_SLIDE, *it, nowX - x, nowY - y, m_pParam);
                }
            }

            // 状態が変化したらコールバックを呼び出す
            if (nowState != (*it)->GetState() && m_pCallback)
            {
                m_pCallback(CONTROL_EVENT_CHANGE_STATE, *it, nowX - x, nowY - y, m_pParam);
            }
        }
        else
        {
            if (m_activeId == (*it)->GetId())
            {
                // 前回のペン位置がコントロール内であったか
                bool isIn = (prevX >= x && prevX <= x + width && prevY >= y && prevY <= y + height);

                // まだタッチされている→ペンスライド
                if (nowTouch)
                {
                    (*it)->OnPenSlide(false);

                    if (m_pCallback)
                    {
                        m_pCallback(CONTROL_EVENT_SLIDE, *it, nowX - x, nowY - y, m_pParam);
                    }
                }
                else
                {
                    (*it)->OnPenUp(isIn);

                    if (m_pCallback)
                    {
                        m_pCallback(CONTROL_EVENT_PENUP, *it, nowX - x, nowY - y, m_pParam);
                    }

                    // コントロール内でペンが離された→ペンタッチ
                    if (isIn)
                    {
                        (*it)->OnPenTouch();

                        if (m_pCallback)
                        {
                            m_pCallback(CONTROL_EVENT_TOUCH, *it, nowX - x, nowY - y, m_pParam);
                        }
                    }

                    m_activeId = CONTROL_ID_UNUSED;
                }

                // 状態が変化したらコールバックを呼び出す
                if (nowState != (*it)->GetState() && m_pCallback)
                {
                    m_pCallback(CONTROL_EVENT_CHANGE_STATE, *it, nowX - x, nowY - y, m_pParam);
                }
            }
        }
    }

    if (nowTouch)
    {
        // nowTouch が有効でないと、 nowX, nowY の値も有効でない
        m_touchStatusOld.x = nowX;
        m_touchStatusOld.y = nowY;
    }

    m_touchStatusOld.touch = nowTouch;
}

// 描画処理を行います。
void ControlManager::Draw()
{
    for (ControlList::iterator it = m_controlList.begin(); it != m_controlList.end(); it++)
    {
        if ((*it)->GetState() != CONTROL_STATE_HIDE)
        {
            (*it)->OnDraw();
        }
    }
}

}
