﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     TextWriter.cpp

  Copyright (C)2009-2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 40635 $
 *---------------------------------------------------------------------------*/
 
#include "TextWriter.h"

//////////////////////////////////////////////////////////////////////////////////////////
// 定数定義
//////////////////////////////////////////////////////////////////////////////////////////

// ディスプレイ数
#define DISPLAY_NUM 2

// ディスプレイ番号
#define DISPLAY_0 0
#define DISPLAY_1 1

extern bit8 NNFONT_RECTDRAWERSHADER_BEGIN[];
extern bit8 NNFONT_RECTDRAWERSHADER_END[];

namespace
{
    // 初期化フラグ
    bool s_isInitialized = false;
    // 単位行列
    nn::math::MTX34 s_mtxIdentify;
    // プロジェクション行列
    nn::math::MTX44 s_mtxProjection[DISPLAY_NUM];

    // カラーバッファ情報
    const nn::font::ColorBufferInfo s_colBufferInfos[DISPLAY_NUM] =
    {
        {NN_GX_DISPLAY0_WIDTH, NN_GX_DISPLAY0_HEIGHT, PICA_DATA_DEPTH24_STENCIL8_EXT},
        {NN_GX_DISPLAY1_WIDTH, NN_GX_DISPLAY1_HEIGHT, PICA_DATA_DEPTH24_STENCIL8_EXT}
    };

    // 描画設定コマンド
    const u32 s_screenSettingCommands0[] =
    {
        // ビューポートの設定
        NN_FONT_CMD_SET_VIEWPORT(0, 0, s_colBufferInfos[DISPLAY_0].width, s_colBufferInfos[DISPLAY_0].height),
        // シザー処理を無効
        NN_FONT_CMD_SET_DISABLE_SCISSOR(s_colBufferInfos[DISPLAY_0]),
        // wバッファの無効化／デプスレンジの設定／ポリゴンオフセットの無効化
        NN_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(0, 0, 1, 0, s_colBufferInfos[DISPLAY_0])
    };

    // 描画設定コマンド
    const u32 s_screenSettingCommands1[] =
    {
        // ビューポートの設定
        NN_FONT_CMD_SET_VIEWPORT(0, 0, s_colBufferInfos[DISPLAY_1].width, s_colBufferInfos[DISPLAY_1].height),
        // シザー処理を無効
        NN_FONT_CMD_SET_DISABLE_SCISSOR(s_colBufferInfos[DISPLAY_1]),
        // wバッファの無効化／デプスレンジの設定／ポリゴンオフセットの無効化
        NN_FONT_CMD_SET_WBUFFER_DEPTHRANGE_POLYGONOFFSET(0, 0, 1, 0, s_colBufferInfos[DISPLAY_1])
    };

    // 初期化コマンド
    const u32 s_initCommands[] =
    {
        // カリングを無効
        NN_FONT_CMD_SET_CULL_FACE(NN_FONT_CMD_CULL_FACE_DISABLE),
        // ステンシルテストを無効
        NN_FONT_CMD_SET_DISABLE_STENCIL_TEST(),
        // デプステストを無効／カラーバッファの全ての成分を書き込み可
        NN_FONT_CMD_SET_DEPTH_FUNC_COLOR_MASK(false, 0, true, true, true, true, true),
        // アーリーデプステストを無効
        NN_FONT_CMD_SET_ENABLE_EARLY_DEPTH_TEST(false),
        // フレームバッファアクセス制御
        NN_FONT_CMD_SET_FBACCESS(true, true, false, false, false, false)
    };
}

//////////////////////////////////////////////////////////////////////////////////////////
namespace scene
{

// コンストラクタです。
TextWriter::TextWriter()
{
    m_pExpHeap      = NULL;
    m_pStringBuffer = NULL;
    m_textColor     = nn::util::Color8(255, 255, 255, 255);
    m_pShaderBuffer = NULL;
    m_targetDisplay = DISPLAY_0;
}

// デストラクタです。
TextWriter::~TextWriter()
{
    Finalize();
}

// 初期化処理を行います。
void TextWriter::Initialize(nn::fnd::ExpHeap* pExpHeap)
{
    // 共有フォントの初期化処理を行う
    if (nn::pl::InitializeSharedFont().IsSuccess())
    {
        // 共有フォントのロードが完了するまで待機する
        while (nn::pl::GetSharedFontLoadState() != nn::pl::SHARED_FONT_LOAD_STATE_LOADED)
        {
            nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(1));
        }

        // 共有フォントのロードに失敗していないか確認する
        if (nn::pl::GetSharedFontLoadState() == nn::pl::SHARED_FONT_LOAD_STATE_FAILED)
        {
            NN_PANIC("nn::pl::InitializeSharedFont() failed!\n");
        }
    }

    m_pExpHeap = pExpHeap;

    // フォントの初期化処理を行う
    if (!InitializeFont())
    {
        NN_PANIC("InitializeFont() failed!\n");
    }

    // シェーダの初期化処理を行う
    if (!InitializeShader())
    {
        NN_PANIC("InitializeShader() failed!\n");
    }

    // 文字列バッファを確保する
    m_pStringBuffer = AllocateStringBuffer(512);

    if (!m_pStringBuffer)
    {
        NN_PANIC("Allocate string buffer failed!\n");
    }

    // 起動中に1回行うだけでよい初期化処理
    if (!s_isInitialized)
    {
        // プロジェクション行列を作成する
        nn::math::MTX44OrthoPivot(&s_mtxProjection[DISPLAY_0], 0, NN_GX_DISPLAY0_HEIGHT, NN_GX_DISPLAY0_WIDTH, 0, 0, -1, nn::math::PIVOT_UPSIDE_TO_TOP);
        nn::math::MTX44OrthoPivot(&s_mtxProjection[DISPLAY_1], 0, NN_GX_DISPLAY1_HEIGHT, NN_GX_DISPLAY1_WIDTH, 0, 0, -1, nn::math::PIVOT_UPSIDE_TO_TOP);

        // 単位行列を作成する
        nn::math::MTX34Identity(&s_mtxIdentify);

        s_isInitialized = true;
    }

    // フォントオブジェクトを設定する
    m_writer.SetFont(&m_font);
    // 初期フォントサイズは大きいので、小さくしておく
    m_writer.SetScale(0.6f);
}

// 終了処理を行います。
void TextWriter::Finalize()
{
    if (m_pExpHeap)
    {
        // 描画用バッファを無効化する
        void* pDrawBuffer = m_font.SetDrawBuffer(NULL);

        if (pDrawBuffer)
        {
            m_pExpHeap->Free(pDrawBuffer);
        }

        if (m_pShaderBuffer)
        {
            m_pExpHeap->Free(m_pShaderBuffer);
        }
        if (m_pStringBuffer)
        {
            m_pExpHeap->Free(m_pStringBuffer);
        }

        // 変数を初期化する
        m_pExpHeap      = NULL;
        m_pStringBuffer = NULL;
        m_pShaderBuffer = NULL;
    }
}

// 描画を開始します。
void TextWriter::BeginRender(GLenum targetDisplay)
{
    if (targetDisplay == NN_GX_DISPLAY0)
    {
        nngxAdd3DCommand(s_screenSettingCommands0, sizeof (s_screenSettingCommands0), true);
        m_targetDisplay = DISPLAY_0;
    }
    else
    {
        nngxAdd3DCommand(s_screenSettingCommands1, sizeof (s_screenSettingCommands1), true);
        m_targetDisplay = DISPLAY_1;
    }

    nngxAdd3DCommand(s_initCommands, sizeof (s_initCommands), true);

    m_writer.SetDispStringBuffer(m_pStringBuffer);
    m_writer.StartPrint();
}

// 描画を終了します。
void TextWriter::EndRender()
{
    m_writer.EndPrint();

    m_drawer.BuildTextCommand(&m_writer);
    m_drawer.DrawBegin();
    {
        // プロジェクション行列を設定する
        m_drawer.SetProjectionMtx(s_mtxProjection[m_targetDisplay]);
        // ビュー行列を初期化する
        m_drawer.SetViewMtxForText(s_mtxIdentify);

        m_writer.UseCommandBuffer();
    }
    m_drawer.DrawEnd();

    // GPUレジスタ設定の復帰処理
    nngxUpdateState(NN_GX_STATE_ALL);
    nngxValidateState(NN_GX_STATE_ALL, GL_TRUE);
}

// 文字列を描画します。
void TextWriter::Printf(f32 x, f32 y, const wchar_t* pFormat, ...)
{
    va_list pArgs;

    va_start(pArgs, pFormat);
    {
        m_writer.SetCursor(x, y);
        m_writer.SetTextColor(m_textColor);
        m_writer.VPrintf(pFormat, pArgs);
    }
    va_end(pArgs);
}

// 文字列を描画します。
void TextWriter::PutString(f32 x, f32 y, const wchar_t* pString)
{
    m_writer.SetCursor(x, y);
    m_writer.SetTextColor(m_textColor);
    m_writer.Print(pString);
}

// 文字色を設定します。
void TextWriter::SetTextColor(u8 red, u8 green, u8 blue, u8 alpha)
{
    // 文字の色
    m_textColor = nn::util::Color8(red, green, blue, alpha);
}

// フォントサイズを取得します。
f32 TextWriter::GetFontSize()
{
    return m_writer.GetFontHeight();
}

// フォントサイズを設定します。
void TextWriter::SetFontSize(f32 size)
{
    m_writer.SetFontSize(size);
}

// フォントの拡大率を取得します。

f32 TextWriter::GetFontScale()
{
    return m_writer.GetScaleH();
}

// フォントの拡大率を設定します。
void TextWriter::SetFontScale(f32 scale)
{
    m_writer.SetScale(scale);
}

// 文字間を取得します。
f32 TextWriter::GetCharSpace()
{
    return m_writer.GetCharSpace();
}

// 文字間を設定します。
void TextWriter::SetCharSpace(f32 charSpace)
{
    m_writer.SetCharSpace(charSpace);
}

// 文字列の幅を計算します。
f32 TextWriter::CalculateStringWidth(const wchar_t* pString)
{
    return m_writer.CalcStringWidth(pString);
}

// 文字列の高さを計算します。
f32 TextWriter::CalculateStringHeight(const wchar_t* pString)
{
    return m_writer.CalcStringHeight(pString);
}

// フォントの初期化処理を行う。
bool TextWriter::InitializeFont()
{
    // フォントバッファ
    void* pBuffer = nn::pl::GetSharedFontAddress();
    // フォントリソースを設定する
    bool isSuccess = m_font.SetResource(pBuffer);

    // 描画用バッファのサイズ
    u32 drawBufferSize = nn::font::ResFont::GetDrawBufferSize(pBuffer);
    // 描画用バッファ
    void* pDrawBuffer = m_pExpHeap->Allocate(drawBufferSize, 4);

    NN_NULL_ASSERT(pDrawBuffer);

    // 描画バッファを設定する
    m_font.SetDrawBuffer(pDrawBuffer);

    return isSuccess;
}

// シェーダの初期化処理を行う。
bool TextWriter::InitializeShader()
{
    const size_t SHADER_SIZE = NNFONT_RECTDRAWERSHADER_END - NNFONT_RECTDRAWERSHADER_BEGIN;
    void* pShaderBinary = m_pExpHeap->Allocate(SHADER_SIZE);

    NN_NULL_ASSERT(pShaderBinary);

    // 頂点用バッファとコマンドバッファのサイズ
    u32 vtxBufferCmdBufferSize = nn::font::RectDrawer::GetVertexBufferCommandBufferSize(NNFONT_RECTDRAWERSHADER_BEGIN, SHADER_SIZE);
    // 頂点用バッファとコマンドバッファ
    m_pShaderBuffer = m_pExpHeap->Allocate(vtxBufferCmdBufferSize);

    NN_NULL_ASSERT(m_pShaderBuffer);

    // 描画オブジェクトの初期化処理を行う
    m_drawer.Initialize(m_pShaderBuffer, NNFONT_RECTDRAWERSHADER_BEGIN, SHADER_SIZE);

    m_pExpHeap->Free(pShaderBinary);

    return true;
}

// 文字列バッファを確保する。
nn::font::DispStringBuffer* TextWriter::AllocateStringBuffer(s32 charMax)
{
    // バッファサイズ
    const u32 bufferSize = nn::font::CharWriter::GetDispStringBufferSize(charMax);
    // バッファ
    void* pBuffer = m_pExpHeap->Allocate(bufferSize);

    NN_NULL_ASSERT(pBuffer);

    return nn::font::CharWriter::InitDispStringBuffer(pBuffer, charMax);
}

} // namespace scene
